//-----------------------------------------------------------------------------
// Totally Ordered Broadcast
//-----------------------------------------------------------------------------

import daj.*;
import java.util.*;


//-----------------------------------------------------------------------------
// the application itself
//-----------------------------------------------------------------------------
public class Main extends Application
{
  public static final int PNUM = 6; // number of nodes
  public static final int DATA_KIND = 0;    // data kind
  public static final int UPDATE_KIND = 1;    // update kind


  //---------------------------------------------------------------------------
  // main function of the application
  //---------------------------------------------------------------------------
  public static void main(String[] args)
    {
      new Main().run();
    }

  //---------------------------------------------------------------------------
  // constructor of application
  //---------------------------------------------------------------------------
  public Main()
    {
      super("Asynchronous Termination Detection", 600, 600);
    }

  //---------------------------------------------------------------------------
  // constrution of the network
  // Every node has a channel to all other nodes
  //---------------------------------------------------------------------------
  public void construct()
    {
      int i, j;
      int[][] pos = {{200, 100}, {400, 100}, {500, 300},
		     {400, 500}, {200, 500}, {100, 300}};
      Node[] nodes = new Node[PNUM];

      for(i = 0; i < PNUM; i++)
	{
	  nodes[i] = node(new Prog(i), 
			  String.valueOf(i), pos[i][0], pos[i][1]);
	}

      for(i=0; i<PNUM; i++)
	{
	  for(j=0; j< PNUM; j++)
	    {
	      if (i != j) link(nodes[i], nodes[j]);
	    }
	}
    }
}
     
//-----------------------------------------------------------------------------
// a program class
//-----------------------------------------------------------------------------
class Prog extends Program
{
  public int nodeID;                   // node number 
  public int N; 
  public MsgQueue msgqueue;  //Queue of data

  //--------------------------------------------------------------------------------------------------------------------------
  // called for the initialization of the program
  //--------------------------------------------------------------------------------------------------------------------------
  public Prog(int process)
    {
      nodeID = process;
      msgqueue = new MsgQueue();
      msgqueue.nodeID = nodeID;
      N = 0;
    }

  //---------------------------------------------------------------------------
  // Totally ordered broadcast send
  //---------------------------------------------------------------------------
  public void bc_send(int data, int kind)
    {
      Msg msg;
      int i;

      switch(kind) {
	case Main.DATA_KIND: 
		msgqueue.ts[nodeID] = msgqueue.ts[nodeID] + 1;
		N = N + 1;
		msg = new Msg(kind,msgqueue.ts[nodeID],N,data,nodeID);
      		msgqueue.AddMsg(msg);
		for(i=0; i<Main.PNUM-1; i++)
		  out(i).send(msg);	
		break;
	case Main.UPDATE_KIND:
		msg = new Msg(kind,data,0,0,nodeID);
		for(i=0; i<Main.PNUM-1; i++)
		  out(i).send(msg);	
		break;
      } //switch
    }


  //---------------------------------------------------------------------------
  // Single Source FIFO receive
  //---------------------------------------------------------------------------
  public boolean bc_recv_ssf()
    {
	int index;

      	index = in().select();
	if (index == -1)
	  return false;
      	Msg msg = (Msg) in(index).receive();
	switch(msg.kind) {
	  case Main.DATA_KIND:
		if (msg.OriginID != nodeID) { // always true
		    msgqueue.ts[msg.OriginID] = msg.T;
		    msgqueue.AddMsg(msg);
		    if (msg.T > msgqueue.ts[nodeID]) {
			msgqueue.ts[nodeID] = msg.T;
			bc_send(msgqueue.ts[nodeID], Main.UPDATE_KIND);
		    }
		}
		break;
	  case Main.UPDATE_KIND:
		if (msg.OriginID != nodeID)  //always true
		    msgqueue.ts[msg.OriginID] = msg.T;
		break;
	} //switch
	return true;
    }


  //---------------------------------------------------------------------------
  // Totally ordered broadcast receive
  //---------------------------------------------------------------------------
  public int bc_recv()
    {
	boolean flag=true;

	while (flag) {
	  Msg msg = (Msg) msgqueue.GetMsg();
	  if (msg == null)
	    bc_recv_ssf();
      	  
	  if (msg != null) {
		flag = false;
		return(msg.Data);
	  }
	}      
        return(-1);
    }



  //---------------------------------------------------------------------------
  // called for the execution of the program
  //---------------------------------------------------------------------------
  public void main()
    {
      int i, value;

      bc_send(nodeID, Main.DATA_KIND);

      if (nodeID == 3)
	bc_send(nodeID+10, Main.DATA_KIND);


      for(i=1; i<=Main.PNUM+1; i++) {
	  value = bc_recv();
	  System.out.println("P"+nodeID+" receive broadcast data from P"+value%10+"("+value/10+")");
      }

      bc_send(nodeID+20, Main.DATA_KIND);


      for(i=1; i<=Main.PNUM; i++) {
	  value = bc_recv();
	  System.out.println("P"+nodeID+" receive broadcast data from P"+value%10+"("+value/10+")");
      }

    }

  //---------------------------------------------------------------------------
  // deliver random number 0 <= r < n
  //---------------------------------------------------------------------------
  Random rand = new Random();
  public int random(int n)
  {
    return (Math.abs(rand.nextInt()) % n);
  }

  //---------------------------------------------------------------------------
  // informative message
  //---------------------------------------------------------------------------
  public String getText()
    {
      return "Node: " + String.valueOf(nodeID);
    }
}

//---------------------------------------------------------------------------------------------------------------------------
// a message class for general purpose messages
//---------------------------------------------------------------------------------------------------------------------------
class Msg extends Message
{
  public int kind; 	// Kind of message
  public int T;		// Number used in totally ordered broadcast
  public int N;		// Message ID in the original process
  public int Data;	// Data in message
  public int OriginID;	// ID of the original process
  public Msg next;	// Point to the next message

  //---------------------------------------------------------------------------
  // constructor of the message object
  //---------------------------------------------------------------------------
  public Msg(int ki,int ti,int ni,int di,int nodeid)
    {
      kind = ki;
      T = ti;
      N = ni;
      Data = di;
      OriginID = nodeid;
    }

  //---------------------------------------------------------------------------
  // get message value
  //---------------------------------------------------------------------------
  public int value()
    {
      return Data;
    }

  //---------------------------------------------------------------------------
  // informative message
  //---------------------------------------------------------------------------
  public String getText()
    {
      return "Msg: " + String.valueOf(Data);
    }
}



//--------------------------------------------------------------------------------------------------------------------------
// a message queue class used to keep pending messages
//--------------------------------------------------------------------------------------------------------------------------
class MsgQueue extends Message
{  
  
  public int[] tn, ts;
  public int nodeID;
  public Msg headMsg;
  

  //---------------------------------------------------------------------------
  // constructor of the message object
  //---------------------------------------------------------------------------
  public MsgQueue()
    {
      headMsg = null;
      ts = new int[Main.PNUM];
      tn = new int[Main.PNUM];
      for(int i=0; i<Main.PNUM; i++) {
	ts[i] = 0;
	tn[i] = 0;
      }
    }

  //---------------------------------------------------------------------------
  // Insert message to the queue
  //---------------------------------------------------------------------------
  public void AddMsg(Msg dmsg)
    {
      Msg tmpMsg, msg;
      Msg frontMsg;

      msg = new Msg(dmsg.kind,dmsg.T,dmsg.N,dmsg.Data,dmsg.OriginID);

      if (headMsg == null) {
	  headMsg = msg;
	  headMsg.next = null;
      } else {
          tmpMsg = headMsg;
	  frontMsg = headMsg;
	  while ((tmpMsg != null) && 
		((tmpMsg.T < msg.T) || ((tmpMsg.T == msg.T) && (tmpMsg.OriginID < msg.OriginID)))) {
	    frontMsg = tmpMsg;
	    tmpMsg = tmpMsg.next;
	  } //while
	  if (tmpMsg == headMsg) {
		msg.next = headMsg;
		headMsg = msg;
	  } else {
		if (tmpMsg == null)
			msg.next = null;
		else
			msg.next = frontMsg.next;
	  	frontMsg.next = msg;

	    }
        } //else
    }


  //---------------------------------------------------------------------------
  // Get message from the queue
  //---------------------------------------------------------------------------
  public Msg GetMsg()
    {
      int i;
      boolean flag;
      Msg tmpMsg;

      if (headMsg == null)
	return null;
      else {
	flag = true;
	i = 0;
	while (flag && (i < Main.PNUM)) {
	  if (headMsg.T > ts[i])
	     flag = false;
	  i++;
	}
	if ((flag) && (headMsg.N == tn[headMsg.OriginID]+1)) {
	  tn[headMsg.OriginID] = tn[headMsg.OriginID] + 1;
	  tmpMsg = headMsg;
	  headMsg = headMsg.next;
	  return(tmpMsg);
	}
	return null;		
      }
    }

  //---------------------------------------------------------------------------
  // informative message
  //---------------------------------------------------------------------------
  public String getText()
    {
      return "Queue: ";
    }
}



//-----------------------------------------------------------------------------
// end of source file
//-----------------------------------------------------------------------------






